# Generated by Django 4.2.27 on 2026-01-06 16:48

from django.db import migrations, models
import django.db.models.deletion
import django_lifecycle.mixins
import pulpcore.app.models.base


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0066_layout_choices'),
    ]

    operations = [
        migrations.CreateModel(
            name='RpmPackageSigningResult',
            fields=[
                ('pulp_id', models.UUIDField(default=pulpcore.app.models.base.pulp_uuid, editable=False, primary_key=True, serialize=False)),
                ('pulp_created', models.DateTimeField(auto_now_add=True)),
                ('pulp_last_updated', models.DateTimeField(auto_now=True, null=True)),
                ('original_package_sha256', models.TextField(max_length=64)),
                ('package_signing_fingerprint', models.TextField(max_length=40)),
                ('result_package', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='core.content')),
            ],
            options={
                'unique_together': {('original_package_sha256', 'package_signing_fingerprint')},
            },
            bases=(django_lifecycle.mixins.LifecycleModelMixin, models.Model),
        ),
    ]
