# Generated by Django 4.2.27 on 2025-12-09 19:34

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0065_alter_package_options'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rpmpublication',
            name='layout',
            field=models.TextField(choices=[('nested_alphabetically', 'nested_alphabetically'), ('flat', 'flat'), ('nested_by_digest', 'nested_by_digest'), ('nested_by_both', 'nested_by_both')], null=True),
        ),
        migrations.AlterField(
            model_name='rpmrepository',
            name='layout',
            field=models.TextField(choices=[('nested_alphabetically', 'nested_alphabetically'), ('flat', 'flat'), ('nested_by_digest', 'nested_by_digest'), ('nested_by_both', 'nested_by_both')], null=True),
        ),
    ]
