# Generated by Django 5.2.5 on 2026-01-29 19:28

from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        ('rpm', '0067_add_rpm_package_signing_result'),
    ]

    operations = [
        migrations.AlterField(
            model_name='rpmpublication',
            name='compression_type',
            field=models.TextField(choices=[('zstd', 'zstd'), ('gz', 'gz'), ('none', 'none')], null=True),
        ),
        migrations.AlterField(
            model_name='rpmrepository',
            name='compression_type',
            field=models.TextField(choices=[('zstd', 'zstd'), ('gz', 'gz'), ('none', 'none')], null=True),
        ),
    ]
